"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.MouseLeaveAllListener = exports.signal = exports.onEditorEvent = exports.fromEditorEvent = void 0;
const codemirror_1 = require("codemirror");
exports.fromEditorEvent = (target, type) => {
    return (cb) => {
        const untupled = (...args) => {
            cb(args);
        };
        target.on(type, untupled);
        return () => {
            target.off(type, untupled);
        };
    };
};
exports.onEditorEvent = (target, type, handler) => exports.fromEditorEvent(target, type)(handler);
/**
 * Typed wrapper around `CodeMirror.signal`.
 * @param editor - Editor instance
 * @param name - Event name
 * @param args - Event parameters
 */
exports.signal = (editor, name, ...args) => {
    codemirror_1.signal(editor, name, ...args);
};
/**
 * Special class for listening to enter/leave events on a group of elements. When none of those elements are
 * being hovered over, a callback is triggered.
 */
class MouseLeaveAllListener {
    /**
     * @param onMouseLeaveAll - Called when none of the tracked elements have a mouse over them.
     */
    constructor(onMouseLeaveAll) {
        this.onMouseLeaveAll = onMouseLeaveAll;
        this.listeners = [];
    }
    /**
     * Adds an element to the list of tracked elements.
     * This element will have mouseenter and mouseleave events tracked.
     * @param element
     */
    addElement(element) {
        if (this.listeners.find((l) => l.element === element))
            return;
        const l = {
            element,
            over: false,
            onEnter: () => {
                l.over = true;
            },
            onLeave: () => {
                l.over = false;
                this.checkHoverState();
            },
        };
        this.listeners.push(l);
        element.addEventListener("mouseenter", l.onEnter);
        element.addEventListener("mouseleave", l.onLeave);
    }
    /**
     * Removes an element from the list of tracked elements.
     * If the element is found, we also remove the event listeners.
     * @param element
     * @return {boolean} - true if the element was found, false otherwise
     */
    removeElement(element) {
        const idx = this.listeners.findIndex((l) => l.element === element);
        const found = idx !== -1;
        if (found) {
            const l = this.listeners[idx];
            this.listeners.splice(idx, 1);
            l.element.removeEventListener("mouseenter", l.onEnter);
            l.element.removeEventListener("mouseleave", l.onLeave);
            this.checkHoverState();
        }
        return found;
    }
    /**
     * Removes all tracked elements from this class.
     */
    dispose() {
        this.listeners.forEach((l) => {
            this.removeElement(l.element);
        });
    }
    checkHoverState() {
        window.setTimeout(() => {
            if (this.listeners.every((el) => !el.over)) {
                this.onMouseLeaveAll();
            }
        }, 500);
    }
}
exports.MouseLeaveAllListener = MouseLeaveAllListener;
//# sourceMappingURL=events.js.map