"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.completionItemsToList = exports.symbolKindToString = exports.completionItemKindToString = exports.hoverContentsToString = exports.documentationToString = exports.diagnosticTagName = exports.diagnosticSeverityName = exports.cmRange = exports.cmPosition = exports.lspChange = exports.lspRange = exports.lspPosition = void 0;
const vscode_languageserver_protocol_1 = require("vscode-languageserver-protocol");
const string_1 = require("./string");
exports.lspPosition = ({ line, ch }) => ({
    line,
    character: ch,
});
exports.lspRange = (start, end) => ({
    start: exports.lspPosition(start),
    end: exports.lspPosition(end),
});
exports.lspChange = (change) => ({
    range: exports.lspRange(change.from, change.to),
    text: change.text.join("\n"),
});
exports.cmPosition = ({ line, character }) => ({
    line,
    ch: character,
});
exports.cmRange = ({ start, end }) => [
    exports.cmPosition(start),
    exports.cmPosition(end),
];
exports.diagnosticSeverityName = (severity) => {
    switch (severity) {
        case 1 /* DiagnosticSeverity.Error */:
            return "error";
        case 2 /* DiagnosticSeverity.Warning */:
            return "warning";
        case 3 /* DiagnosticSeverity.Information */:
            return "information";
        case 4 /* DiagnosticSeverity.Hint */:
            return "hint";
    }
};
exports.diagnosticTagName = (tag) => {
    switch (tag) {
        case 1 /* DiagnosticTag.Unnecessary */:
            // Style this with faded text.
            return "unnecessary";
        case 2 /* DiagnosticTag.Deprecated */:
            // Style this with strike through.
            return "deprecated";
    }
};
exports.documentationToString = (doc, renderMarkdown = (x) => x) => {
    if (!doc)
        return "";
    if (typeof doc === "string")
        return string_1.escapeHtml(doc);
    return markupContentToString(doc, renderMarkdown);
};
exports.hoverContentsToString = (contents, renderMarkdown) => {
    // Note that MarkedString is considereed deprecated
    // MarkedString[]
    if (Array.isArray(contents)) {
        return contents
            .map((c) => markedStringToString(c, renderMarkdown))
            .join("\n");
    }
    if (vscode_languageserver_protocol_1.MarkedString.is(contents)) {
        return markedStringToString(contents, renderMarkdown);
    }
    return markupContentToString(contents, renderMarkdown);
};
const markupContentToString = (content, renderMarkdown) => content.kind === "markdown"
    ? renderMarkdown(content.value)
    : string_1.escapeHtml(content.value);
const markedStringToString = (m, render) => {
    // Remove two backslashes in front of periods and hyphens in plain text.
    if (typeof m === "string")
        return m.replace(/\\(?=[-.])/g, "");
    // To Markdown code block
    return render(["```" + m.language, m.value, "```"].join("\n"));
};
exports.completionItemKindToString = (() => {
    const kinds = [
        "",
        "text",
        "method",
        "function",
        "constructor",
        "field",
        "variable",
        "class",
        "interface",
        "module",
        "property",
        "unit",
        "value",
        "enum",
        "keyword",
        "snippet",
        "color",
        "file",
        "reference",
        "folder",
        "enum-member",
        "constant",
        "struct",
        "event",
        "operator",
        "type-parameter",
    ];
    return (kind) => kinds[kind !== null && kind !== void 0 ? kind : 0] || "unknown";
})();
exports.symbolKindToString = (() => {
    const kinds = [
        "",
        "file",
        "module",
        "namespace",
        "package",
        "class",
        "method",
        "property",
        "field",
        "constructor",
        "enum",
        "interface",
        "function",
        "variable",
        "constant",
        "string",
        "number",
        "boolean",
        "array",
        "object",
        "key",
        "null",
        "enum-member",
        "struct",
        "event",
        "operator",
        "type-parameter",
    ];
    return (kind) => kinds[kind !== null && kind !== void 0 ? kind : 0] || "unknown";
})();
// > If a `CompletionItem[]` is provided it is interpreted to be complete.
// > So it is the same as `{ isIncomplete: false, items }`
exports.completionItemsToList = (items) => ({
    items,
    isIncomplete: false,
});
//# sourceMappingURL=conversions.js.map