"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Workspace = void 0;
const codemirror_1 = require("codemirror");
const vscode_languageserver_protocol_1 = require("vscode-languageserver-protocol");
const vscode_jsonrpc_ws_1 = require("@qualified/vscode-jsonrpc-ws");
const vscode_jsonrpc_ww_1 = require("@qualified/vscode-jsonrpc-ww");
const lsp_connection_1 = require("@qualified/lsp-connection");
const capabilities_1 = require("./capabilities");
const event_stream_1 = require("./utils/event-stream");
const events_1 = require("./events");
const conversions_1 = require("./utils/conversions");
const editor_1 = require("./utils/editor");
const promise_1 = require("./utils/promise");
const context_menu_1 = require("./ui/context-menu");
// Changes stream emits at most once per 50ms.
const CHANGES_FRAME = 50;
/**
 * Workspace provides code intelligence to CodeMirror editors by managing
 * communications with Language Servers and adding event handlers.
 */
class Workspace {
    /**
     * Create new workspace.
     * @param options
     */
    constructor(options) {
        this.editors = Object.create(null);
        this.connectPromises = Object.create(null);
        this.connections = Object.create(null);
        this.documentVersions = Object.create(null);
        this.subscriptionDisposers = new WeakMap();
        this.completionHandlers = new WeakMap();
        this.configs = options.configs || {};
        this.rootUri =
            options.rootUri + (!options.rootUri.endsWith("/") ? "/" : "");
        this.getConnectionString = options.getConnectionString.bind(void 0);
        this.getLanguageAssociation = options.getLanguageAssociation.bind(void 0);
        this.canHandleMarkdown = typeof options.renderMarkdown === "function";
        const renderMarkdown = options.renderMarkdown || ((x) => x);
        this.renderMarkdown = renderMarkdown.bind(void 0);
    }
    /**
     * Dispose the workspace.
     * Close connections and remove references to editors.
     */
    dispose() {
        // TODO shutdown and exit all connections
    }
    /**
     * Open text document in the workspace to notify the Language Server and
     * enable code intelligence.
     * @param path - The file path relative to the project root.
     * @param editor - CodeMirror Editor instance.
     */
    async openTextDocument(path, editor) {
        const uri = this.getDocumentUri(path);
        const assoc = this.getLanguageAssociation(uri);
        if (!assoc)
            return;
        // TODO Allow connecting to multiple language servers
        const serverId = assoc.languageServerIds[0];
        if (!serverId)
            return;
        const conn = await this.getConnection(serverId);
        if (!conn)
            return;
        this.editors[uri] = editor;
        this.documentVersions[uri] = 0;
        const languageId = assoc.languageId;
        conn.textDocumentOpened({
            textDocument: {
                uri,
                languageId,
                text: editor.getValue(),
                version: ++this.documentVersions[uri],
            },
        });
        this.addEventHandlers(uri, editor, conn);
    }
    // TODO Clean up. Workspace should signal custom events for providers to react
    addEventHandlers(uri, editor, conn) {
        const completion = new capabilities_1.CompletionHandler({
            uri,
            editor,
            conn,
            renderMarkdown: this.renderMarkdown,
        });
        this.completionHandlers.set(editor, completion);
        const disposers = [];
        // create a listener to track whether the mouse is over the editor and/or any of it's tooltips
        const mouseLeaveAllListener = new events_1.MouseLeaveAllListener(() => {
            this.hideTooltips(editor);
        });
        mouseLeaveAllListener.addElement(editor.getWrapperElement());
        disposers.push(() => {
            mouseLeaveAllListener.dispose();
        });
        const requestCompletion = (cm) => {
            capabilities_1.removeSignatureHelp(cm);
            completion.invoke();
        };
        const changeStream = event_stream_1.piped(events_1.fromEditorEvent(editor, "changes"), event_stream_1.debouncedBuffer(CHANGES_FRAME), event_stream_1.map((buffered) => {
            const cm = buffered[0][0];
            // Send incremental contentChanges
            conn.textDocumentChanged({
                textDocument: {
                    uri,
                    version: ++this.documentVersions[uri],
                },
                contentChanges: conn.syncsIncrementally
                    ? buffered.flatMap(([_, cs]) => cs.map(conversions_1.lspChange))
                    : [{ text: cm.getValue() }],
            });
            // Only pass the editor and the last change object.
            const lastChanges = buffered[buffered.length - 1][1];
            return [cm, lastChanges[lastChanges.length - 1]];
        }), event_stream_1.filter(([cm, change]) => {
            // Text removed
            if (change.origin === "+delete" ||
                change.text.every((s) => s.length === 0)) {
                completion.close();
                capabilities_1.removeSignatureHelp(cm);
                return false;
            }
            // Text completed
            if (change.origin === "complete" || change.origin === "+complete") {
                return false;
            }
            return true;
        }));
        disposers.push(
        // Trigger completion or signature help or hide them
        changeStream(([cm, _change]) => {
            if (completion.onChange())
                return;
            const pos = cm.getCursor();
            if (pos.ch === 0)
                return;
            const triggerCharacter = cm.getLine(pos.line)[pos.ch - 1];
            if (conn.signatureHelpTriggers.includes(triggerCharacter) ||
                conn.signatureHelpRetriggers.includes(triggerCharacter)) {
                completion.close();
                capabilities_1.removeSignatureHelp(cm);
                conn
                    .getSignatureHelp({
                    textDocument: { uri },
                    position: conversions_1.lspPosition(pos),
                    context: {
                        triggerKind: vscode_languageserver_protocol_1.SignatureHelpTriggerKind.TriggerCharacter,
                        triggerCharacter,
                        // TODO Look into this
                        isRetrigger: false,
                    },
                })
                    .then((help) => {
                    if (!help || help.signatures.length === 0)
                        return;
                    capabilities_1.showSignatureHelp(cm, mouseLeaveAllListener, help, pos, this.renderMarkdown);
                });
                return;
            }
            completion.close();
            capabilities_1.removeSignatureHelp(cm);
        }));
        // Highlights identifiers matching the word under cursor
        // Note that `cursorActivity` is also emitted when content changes.
        const cursorActivityStream = event_stream_1.piped(events_1.fromEditorEvent(editor, "cursorActivity"), event_stream_1.debounce(100), event_stream_1.map(([cm]) => [cm, cm.getCursor()]));
        disposers.push(cursorActivityStream(([cm, pos]) => {
            const token = cm.getTokenAt(pos);
            if (token.type === "variable" || token.type === "property") {
                conn
                    .getDocumentHighlight({
                    textDocument: { uri },
                    position: conversions_1.lspPosition(pos),
                })
                    .then((highlights) => {
                    capabilities_1.removeHighlights(cm);
                    if (highlights)
                        capabilities_1.showHighlights(cm, highlights);
                });
            }
            else {
                capabilities_1.removeHighlights(cm);
                if (pos.ch > 0) {
                    const triggerCharacter = cm.getLine(pos.line)[pos.ch - 1];
                    if (!conn.signatureHelpTriggers.includes(triggerCharacter) &&
                        !conn.signatureHelpRetriggers.includes(triggerCharacter)) {
                        capabilities_1.removeSignatureHelp(cm);
                    }
                }
                else {
                    capabilities_1.removeSignatureHelp(cm);
                }
            }
        }));
        // Show hover information on mouseover
        const mouseoverStream = event_stream_1.piped(event_stream_1.fromDomEvent(editor.getWrapperElement(), "mouseover"), event_stream_1.filter(() => capabilities_1.hoverInfoEnabled(editor)), event_stream_1.debounce(100), event_stream_1.map((ev) => editor.coordsChar({ left: ev.pageX, top: ev.pageY }, "page")), 
        // Ignore same position
        event_stream_1.skipDuplicates((p1, p2) => {
            if (p1.line !== p2.line)
                return false;
            if (p1.ch === p2.ch)
                return true;
            const t1 = editor.getTokenAt(p1);
            const t2 = editor.getTokenAt(p2);
            return (t1.string === t2.string && t1.start === t2.start && t1.end === t2.end);
        }));
        disposers.push(mouseoverStream((pos) => {
            capabilities_1.removeHoverInfo(editor);
            const token = editor.getTokenAt(pos);
            if (token.type === "comment" ||
                token.string.length === 0 ||
                token.type === null) {
                return;
            }
            conn
                .getHoverInfo({
                textDocument: { uri },
                position: conversions_1.lspPosition(pos),
            })
                .then((hover) => {
                if (hover) {
                    capabilities_1.removeSignatureHelp(editor);
                    capabilities_1.showHoverInfo(editor, mouseLeaveAllListener, pos, hover, this.renderMarkdown);
                }
            });
        }));
        const hideAllPopups = ([cm]) => {
            this.hideAllPopups(cm);
        };
        const hideTooltips = ([cm]) => {
            this.hideTooltips(cm);
        };
        disposers.push(events_1.onEditorEvent(editor, "cmw:contextMenuOpened", ([cm]) => {
            capabilities_1.disableHoverInfo(cm);
            completion.close();
            capabilities_1.removeSignatureHelp(cm);
        }), events_1.onEditorEvent(editor, "cmw:contextMenuClosed", ([cm]) => {
            capabilities_1.enableHoverInfo(cm);
        }), 
        // if the editor loses focus, or gets completely reset, hide all overlays, because they are no longer valid
        events_1.onEditorEvent(editor, "blur", hideAllPopups), events_1.onEditorEvent(editor, "refresh", hideAllPopups), 
        // if we type something, or modify the viewport, hide any tooltips, because they no longer line up correctly
        events_1.onEditorEvent(editor, "changes", hideTooltips), events_1.onEditorEvent(editor, "viewportChange", hideTooltips));
        const gotoDefinition = (cm, pos = cm.getCursor()) => {
            conn
                .getDefinition({
                textDocument: { uri },
                position: conversions_1.lspPosition(pos),
            })
                .then((location) => {
                if (location)
                    capabilities_1.gotoLocation(cm, uri, location);
            });
        };
        const gotoDeclaration = (cm, pos = cm.getCursor()) => {
            conn
                .getDeclaration({
                textDocument: { uri },
                position: conversions_1.lspPosition(pos),
            })
                .then((location) => {
                if (location)
                    capabilities_1.gotoLocation(cm, uri, location);
            });
        };
        const gotoTypeDefinition = (cm, pos = cm.getCursor()) => {
            conn
                .getTypeDefinition({
                textDocument: { uri },
                position: conversions_1.lspPosition(pos),
            })
                .then((location) => {
                if (location)
                    capabilities_1.gotoLocation(cm, uri, location);
            });
        };
        const gotoReferences = (cm, pos = cm.getCursor()) => {
            conn
                .getReferences({
                textDocument: { uri },
                position: conversions_1.lspPosition(pos),
                context: {
                    includeDeclaration: true,
                },
            })
                .then((location) => {
                if (location)
                    capabilities_1.gotoLocation(cm, uri, location);
            });
        };
        const gotoImplementations = (cm, pos = cm.getCursor()) => {
            conn
                .getImplementation({
                textDocument: { uri },
                position: conversions_1.lspPosition(pos),
            })
                .then((location) => {
                if (location)
                    capabilities_1.gotoLocation(cm, uri, location);
            });
        };
        disposers.push(events_1.onEditorEvent(editor, "contextmenu", ([cm, e]) => {
            e.preventDefault();
            const pos = cm.coordsChar({ left: e.pageX, top: e.pageY }, "page");
            // TODO Disable items if the server doesn't support it.
            context_menu_1.showContextMenu(cm, e.pageX, e.pageY, [
                [
                    {
                        label: "Go to Definition",
                        handler: () => {
                            gotoDefinition(cm, pos);
                        },
                    },
                    {
                        label: "Go to Type Definition",
                        handler: () => {
                            gotoTypeDefinition(cm, pos);
                        },
                    },
                    {
                        label: "Go to Implementations",
                        handler: () => {
                            gotoImplementations(cm, pos);
                        },
                    },
                    {
                        label: "Go to References",
                        handler: () => {
                            gotoReferences(cm, pos);
                        },
                    },
                    {
                        label: "Go to Symbol...",
                        handler: () => {
                            conn
                                .getDocumentSymbol({
                                textDocument: { uri },
                            })
                                .then((symbols) => {
                                if (symbols)
                                    capabilities_1.showSymbolSelector(cm, uri, symbols);
                            });
                        },
                    },
                ],
                // TODO Handle Copy and Cut because we won't show the browser's context menu.
                // Paste requires explicit permission.
                [
                    {
                        label: "Copy",
                    },
                    {
                        label: "Cut",
                    },
                ],
            ]);
        }));
        // Add some keymaps for jumping to various locations.
        const keyMap = codemirror_1.normalizeKeyMap({
            // TODO Make this configurable
            "Ctrl-Space": (cm) => {
                requestCompletion(cm);
            },
            "Alt-G D": (cm) => {
                gotoDefinition(cm);
            },
            "Alt-G H": (cm) => {
                gotoDeclaration(cm);
            },
            "Alt-G T": (cm) => {
                gotoTypeDefinition(cm);
            },
            "Alt-G I": (cm) => {
                gotoImplementations(cm);
            },
            "Alt-G R": (cm) => {
                gotoReferences(cm);
            },
        });
        editor.addKeyMap(keyMap);
        disposers.push(() => {
            editor.removeKeyMap(keyMap);
        });
        this.subscriptionDisposers.set(editor, disposers);
    }
    hideAllPopups(cm) {
        if (cm) {
            this.hideTooltips(cm);
            const completion = this.completionHandlers.get(cm);
            if (completion)
                completion.close();
            context_menu_1.removeContextMenu(cm);
        }
        else {
            Object.values(this.editors).forEach((cm) => this.hideAllPopups(cm));
        }
    }
    hideTooltips(cm) {
        if (cm) {
            capabilities_1.removeHoverInfo(cm);
            capabilities_1.removeSignatureHelp(cm);
        }
        else {
            Object.values(this.editors).forEach((cm) => this.hideTooltips(cm));
        }
    }
    /**
     * Close text document in the workspace to notify the Language Server and
     * remove everything added by the workspace.
     * @param path - The file path relative to the project root.
     */
    async closeTextDocument(path) {
        const uri = this.getDocumentUri(path);
        const assoc = this.getLanguageAssociation(uri);
        if (!assoc)
            return;
        const serverId = assoc.languageServerIds[0];
        if (!serverId)
            return;
        const conn = this.connections[serverId];
        if (!conn)
            return;
        const editor = this.editors[uri];
        delete this.editors[uri];
        delete this.documentVersions[uri];
        this.removeEventHandlers(editor);
        conn.textDocumentClosed({
            textDocument: { uri },
        });
    }
    /**
     * Notify the Language Server that the text document was saved.
     *
     * If connected to a WebSocket proxy with synchronization enabled,
     * the contents of the file will be written to disk.
     * @param path - The file path relative to the project root.
     */
    async saveTextDocument(path) {
        const uri = this.getDocumentUri(path);
        // TODO Support `willSave` with `reason` and `willSaveWaitUntil`
        const assoc = this.getLanguageAssociation(uri);
        if (!assoc)
            return;
        const serverId = assoc.languageServerIds[0];
        if (!serverId)
            return;
        const conn = this.connections[serverId];
        if (!conn)
            return;
        const editor = this.editors[uri];
        if (!editor)
            return;
        // TODO Find Language Server supporting these to test
        conn.textDocumentWillSave({
            textDocument: { uri },
            reason: vscode_languageserver_protocol_1.TextDocumentSaveReason.Manual,
        });
        const edits = await conn.getEditsBeforeSave({
            textDocument: { uri },
            reason: vscode_languageserver_protocol_1.TextDocumentSaveReason.Manual,
        });
        if (edits) {
            editor_1.applyEdits(editor, edits, "beforeSave");
            await promise_1.delay(CHANGES_FRAME * 1.5);
        }
        conn.textDocumentSaved({
            textDocument: { uri },
            text: editor.getValue(),
        });
    }
    /**
     * Send `workspace/didChangeWatchedFiles` notification to all active Language Servers.
     *
     * Note that `FileEvent` must have a valid `uri` starting with `rootUri`.
     * @param changes
     */
    async notifyFilesChanged(changes) {
        for (const conn of Object.values(this.connections)) {
            conn.watchedFilesChanged({ changes });
        }
    }
    async getConnection(serverId) {
        return (this.connectPromises[serverId] ||
            (this.connectPromises[serverId] = this.connect(serverId)));
    }
    removeEventHandlers(editor) {
        const disposers = this.subscriptionDisposers.get(editor);
        if (disposers) {
            for (const dispose of disposers)
                dispose();
            disposers.length = 0;
            this.subscriptionDisposers.delete(editor);
        }
        capabilities_1.removeDiagnostics(editor);
        capabilities_1.removeHoverInfo(editor);
        capabilities_1.removeHighlights(editor);
        capabilities_1.removeSignatureHelp(editor);
        const completion = this.completionHandlers.get(editor);
        if (completion) {
            completion.close();
            this.completionHandlers.delete(editor);
        }
    }
    /**
     * Private method to connect to the language server if possible.
     * If existing connection exists, it'll be shared.
     *
     * @param serverId - ID of the language server.
     */
    async connect(serverId) {
        var _a;
        const existing = this.connections[serverId];
        if (existing || existing === null)
            return existing;
        const connectionString = await this.getConnectionString(serverId);
        if (!connectionString)
            return null;
        // If we got some string that doesn't start with Web Socket protocol, assume
        // it's the worker's location.
        const messageConn = /^wss?:\/\//.test(connectionString)
            ? vscode_jsonrpc_ws_1.createMessageConnection(new WebSocket(connectionString))
            : vscode_jsonrpc_ww_1.createMessageConnection(new Worker(connectionString));
        const conn = await messageConn.then(lsp_connection_1.createLspConnection);
        this.connections[serverId] = conn;
        conn.onClose(() => {
            delete this.connections[serverId];
            delete this.connectPromises[serverId];
        });
        conn.listen();
        const config = this.configs[serverId];
        await conn.initialize({
            capabilities: {
                textDocument: {
                    synchronization: {
                        dynamicRegistration: true,
                        willSave: true,
                        willSaveWaitUntil: true,
                        didSave: true,
                    },
                    completion: {
                        dynamicRegistration: true,
                        completionItem: {
                            snippetSupport: true,
                            insertReplaceSupport: true,
                            // TODO Commit characters cannot be supported with show-hint addon because by the time the
                            //      commit character is detected, the Widget is updated and selected item is gone.
                            //      Maybe assign it somewhere inside `select` event.
                            commitCharactersSupport: false,
                            documentationFormat: this.canHandleMarkdown
                                ? ["markdown", "plaintext"]
                                : ["plaintext"],
                            deprecatedSupport: true,
                            preselectSupport: true,
                            resolveSupport: { properties: ["documentation", "detail"] },
                        },
                        contextSupport: true,
                    },
                    hover: {
                        dynamicRegistration: true,
                        contentFormat: this.canHandleMarkdown
                            ? ["markdown", "plaintext"]
                            : ["plaintext"],
                    },
                    signatureHelp: {
                        dynamicRegistration: true,
                        signatureInformation: {
                            documentationFormat: this.canHandleMarkdown
                                ? ["markdown", "plaintext"]
                                : ["plaintext"],
                            parameterInformation: {
                                labelOffsetSupport: true,
                            },
                        },
                        contextSupport: true,
                    },
                    declaration: {
                        dynamicRegistration: true,
                        linkSupport: false,
                    },
                    definition: {
                        dynamicRegistration: true,
                        linkSupport: true,
                    },
                    typeDefinition: {
                        dynamicRegistration: true,
                        linkSupport: true,
                    },
                    implementation: {
                        dynamicRegistration: true,
                        linkSupport: true,
                    },
                    references: {
                        dynamicRegistration: true,
                    },
                    documentHighlight: {
                        dynamicRegistration: true,
                    },
                    documentSymbol: {
                        dynamicRegistration: true,
                        hierarchicalDocumentSymbolSupport: true,
                    },
                    // codeAction: {},
                    // codeLens: {},
                    // documentLink: {
                    //   dynamicRegistration: true,
                    //   tooltipSupport: false,
                    // },
                    // colorProvider: {},
                    // formatting: {},
                    // rangeFormatting: {},
                    // onTypeFormatting: {},
                    // rename: {},
                    // foldingRange: {},
                    // selectionRange: {},
                    publishDiagnostics: {
                        relatedInformation: true,
                        tagSupport: {
                            valueSet: [vscode_languageserver_protocol_1.DiagnosticTag.Unnecessary, vscode_languageserver_protocol_1.DiagnosticTag.Deprecated],
                        },
                    },
                    moniker: {},
                },
                workspace: {
                    didChangeConfiguration: {
                        dynamicRegistration: true,
                    },
                    didChangeWatchedFiles: {
                        dynamicRegistration: false,
                    },
                },
            },
            // clientInfo: { name: "codemirror-workspace" },
            initializationOptions: (_a = config === null || config === void 0 ? void 0 : config.initOptions) !== null && _a !== void 0 ? _a : null,
            processId: null,
            rootUri: this.rootUri,
            workspaceFolders: null,
        });
        conn.initialized();
        const settings = config === null || config === void 0 ? void 0 : config.settings;
        if (settings)
            conn.configurationChanged({ settings });
        // Add event handlers to pass payload to matching open editors.
        conn.onDiagnostics(({ uri, diagnostics }) => {
            const editor = this.editors[uri];
            if (editor)
                capabilities_1.showDiagnostics(editor, diagnostics);
        });
        return conn;
    }
    getDocumentUri(path) {
        return this.rootUri + path.replace(/^\/+/, "");
    }
}
exports.Workspace = Workspace;
// Renaming file should be done by:
// 1. Close
// 2. Delete
// 3. Create
// 4. Open
//# sourceMappingURL=workspace.js.map