import { MarkedString } from "vscode-languageserver-protocol";
import { escapeHtml } from "./string";
export const lspPosition = ({ line, ch }) => ({
    line,
    character: ch,
});
export const lspRange = (start, end) => ({
    start: lspPosition(start),
    end: lspPosition(end),
});
export const lspChange = (change) => ({
    range: lspRange(change.from, change.to),
    text: change.text.join("\n"),
});
export const cmPosition = ({ line, character }) => ({
    line,
    ch: character,
});
export const cmRange = ({ start, end }) => [
    cmPosition(start),
    cmPosition(end),
];
export const diagnosticSeverityName = (severity) => {
    switch (severity) {
        case 1 /* DiagnosticSeverity.Error */:
            return "error";
        case 2 /* DiagnosticSeverity.Warning */:
            return "warning";
        case 3 /* DiagnosticSeverity.Information */:
            return "information";
        case 4 /* DiagnosticSeverity.Hint */:
            return "hint";
    }
};
export const diagnosticTagName = (tag) => {
    switch (tag) {
        case 1 /* DiagnosticTag.Unnecessary */:
            // Style this with faded text.
            return "unnecessary";
        case 2 /* DiagnosticTag.Deprecated */:
            // Style this with strike through.
            return "deprecated";
    }
};
export const documentationToString = (doc, renderMarkdown = (x) => x) => {
    if (!doc)
        return "";
    if (typeof doc === "string")
        return escapeHtml(doc);
    return markupContentToString(doc, renderMarkdown);
};
export const hoverContentsToString = (contents, renderMarkdown) => {
    // Note that MarkedString is considereed deprecated
    // MarkedString[]
    if (Array.isArray(contents)) {
        return contents
            .map((c) => markedStringToString(c, renderMarkdown))
            .join("\n");
    }
    if (MarkedString.is(contents)) {
        return markedStringToString(contents, renderMarkdown);
    }
    return markupContentToString(contents, renderMarkdown);
};
const markupContentToString = (content, renderMarkdown) => content.kind === "markdown"
    ? renderMarkdown(content.value)
    : escapeHtml(content.value);
const markedStringToString = (m, render) => {
    // Remove two backslashes in front of periods and hyphens in plain text.
    if (typeof m === "string")
        return m.replace(/\\(?=[-.])/g, "");
    // To Markdown code block
    return render(["```" + m.language, m.value, "```"].join("\n"));
};
export const completionItemKindToString = (() => {
    const kinds = [
        "",
        "text",
        "method",
        "function",
        "constructor",
        "field",
        "variable",
        "class",
        "interface",
        "module",
        "property",
        "unit",
        "value",
        "enum",
        "keyword",
        "snippet",
        "color",
        "file",
        "reference",
        "folder",
        "enum-member",
        "constant",
        "struct",
        "event",
        "operator",
        "type-parameter",
    ];
    return (kind) => kinds[kind !== null && kind !== void 0 ? kind : 0] || "unknown";
})();
export const symbolKindToString = (() => {
    const kinds = [
        "",
        "file",
        "module",
        "namespace",
        "package",
        "class",
        "method",
        "property",
        "field",
        "constructor",
        "enum",
        "interface",
        "function",
        "variable",
        "constant",
        "string",
        "number",
        "boolean",
        "array",
        "object",
        "key",
        "null",
        "enum-member",
        "struct",
        "event",
        "operator",
        "type-parameter",
    ];
    return (kind) => kinds[kind !== null && kind !== void 0 ? kind : 0] || "unknown";
})();
// > If a `CompletionItem[]` is provided it is interpreted to be complete.
// > So it is the same as `{ isIncomplete: false, items }`
export const completionItemsToList = (items) => ({
    items,
    isIncomplete: false,
});
//# sourceMappingURL=conversions.js.map