// Creating Streams
export const fromDomEvent = (target, type) => {
    return (cb) => {
        target.addEventListener(type, cb);
        return () => {
            target.removeEventListener(type, cb);
        };
    };
};
// Operators
// Transform value
export const map = (transform) => (stream) => {
    return (cb) => stream((x) => cb(transform(x)));
};
// Perform side effect
export const tap = (effect) => (stream) => {
    return (cb) => {
        return stream((x) => {
            effect(x);
            cb(x);
        });
    };
};
// stream: ____1____2____2____3
// result: ____1____2_________3
export const skipDuplicates = (equals = (a, b) => a === b) => (stream) => {
    return (cb) => {
        let prev = null;
        return stream((x) => {
            if (!prev || !equals(prev.value, x)) {
                prev = { value: x };
                cb(x);
            }
        });
    };
};
export const debounce = (ms) => (stream) => {
    return (cb) => {
        let timeout = null;
        return stream((x) => {
            if (timeout !== null) {
                clearTimeout(timeout);
                timeout = null;
            }
            timeout = setTimeout(() => cb(x), ms);
        });
    };
};
export const debouncedBuffer = (ms) => (stream) => {
    return (cb) => {
        let timeout = null;
        const buffer = [];
        return stream((x) => {
            if (timeout !== null) {
                clearTimeout(timeout);
                timeout = null;
            }
            buffer.push(x);
            timeout = setTimeout(() => {
                cb(buffer.slice());
                buffer.length = 0;
            }, ms);
        });
    };
};
export const filter = (test) => (stream) => {
    return (cb) => {
        return stream((x) => {
            if (test(x))
                cb(x);
        });
    };
};
// For returning a Stream piped through operators.
export const piped = (stream, ...ops) => ops.reduce((v, op) => op(v), stream);
//# sourceMappingURL=event-stream.js.map